#ifndef _AES_UMX_AES_HXX
#define _AES_UMX_AES_HXX

/*
 * ATAES 132 control functions
 */

#include <sstream>	
#include "umi/uc_d.hxx"
#include "umi/umi_task.hxx"
#include "aesc/umipro_aes.hxx"


namespace INdepAes100 {	
  namespace Umi=INdepUmi100 ;

  extern const Aes_aes_Proto theAes_aes_Proto ;



  class AesTaskPlex:public Umi::TaskPlex,  public Failables {
    UmiChanUser *_pucIo;
    Umi::Channel *_puc;

    bool _fRun;
    std::string _sgIdent;		/* hardware id string */

  protected:
    void done(Umi::Task *pta);
    void run(Umi::TaskFunc::Func *ptf);

   public:
    bool isAesOpen() { return _sgIdent.size(); }
    int eModel();
    bool isModel(int e) { return eModel() == e ; }


    const std::string &sgIdent() const { return _sgIdent; }


    /** open and get the AES id string, _sgIdent */
    Failables &callAesOpen(); 

    /** execute AES command */
    Failable callAesCmd(std::streambuf *psbRecv, std::streambuf *psbSend);
    
    AesTaskPlex(Umi::Channel *puc, Umi::ChanUser *pucIo):Umi::TaskPlex(puc), 
      _pucIo(pucIo), 
      _puc(puc), 
      _fRun(false)
    {
      assert(*(pucIo->pup()) == theAes_aes_Proto); /* REMIND should be >= operator */
      assert(*(puc->pup()) == theAes_aes_Proto); 
    }
    ~AesTaskPlex() { }
  };

  extern Failable readConfig(std::string *psg, AesTaskPlex *prp);
  extern int checkAuth(AesTaskPlex *prp, int iKey);

  class AesLocalTaskInitializer:public Failables {
    Umi::ChanDevio *_puc;	/* szDevice i/o channel */
    AesTaskPlex *_pap;		/* endpoint plexer */
   public:
    AesTaskPlex *pap() const { return _pap; }

    /**
     * configure all resources necessary for process-local operation
     *
     * \arg szDevice system specific path to device node, e.g. "/dev/itina/0"
     *
     * On return check via isGood() for successful initialization.
     */
    AesLocalTaskInitializer(const char *szDevice);
    ~AesLocalTaskInitializer();
  };
}



#endif
