#ifndef _DINO_DINO_HXX_
#define _DINO_DINO_HXX_

#include <map>
#include <sstream>
#include <vector>
#include "umi/umi_task.hxx"
#include "dino/umipro_dino.hxx"
#include "cplRaw/sm.hxx"

/**
 * \file 
 * ines Digital I/O (Dino) interface 
 */
namespace INdepUmi100 {
  class RunIo ;
  class UseIo {
    RunIo* _pur;
   public:

    RunIo* pur() { return _pur; }
    UseIo(RunIo *pur):_pur(pur) { }
    ~UseIo() { }
  };
  class RunIo {
    UseIo *_pui;
    class Chan:public SmM {
      ChanUser *_pcu;
     public:
      ChanUser *pcu() { return _pcu; }
      Chan(ChanUser *pcu=0):_pcu(pcu) {}
      ~Chan() {}
    };
    std::vector<Chan> _vccn;
    typedef std::vector<Chan>::iterator ItCn;
    enum { sIdle, sSend, sRecv };
   public:
    ChanUser *pucIo() { return _vccn.at(0).pcu() ; }

    int run();
    void init(UseIo *pui) { _pui=pui; }
    void add(ChanUser *pucIo) {
      _vccn.push_back(Chan(pucIo));
    }
    RunIo():_pui(0) {  }
  };
}

/** namespace for the ines Digital I/O (Dino) interface */
namespace INdepDino100 {
  namespace Umi=INdepUmi100 ;
  const Dino_dino_Proto theDino_dino_Proto ;

  /**
   * A BitHandle is a compact (integer-type) representation of a bit at a port.
   */
  typedef int BitHandle;
  inline BitHandle bitHandle(int iBits, int iIndex, int f) { return ((iBits << 24) | (iIndex << 16) | f); }
  inline int ibits(BitHandle bh) { return bh >> 24; }
  inline int index(BitHandle bh) { return (bh >> 16) & 0xff; }
  inline unsigned char flag(BitHandle bh) { return bh & 0xff; }
  enum { bhNil = 0 };	/* not a BitHandle */

  /* 
   * Multiple named channel client plexer 
   */
  class MultiPlex : protected std::vector<Umi::TaskFunc> , public Umi::Plexer, public Umi::TaskUser {
    const Umi::Proto *_pup;
    std::map<std::string, unsigned> _mpsi;
    typedef std::map<std::string, unsigned>::iterator ItSi;

    /*
     * find port, nonzero if unknown
     */
   protected:
    int findPort(std::string *psgPort, const std::string &sgPath);
    Umi::TaskFunc *pta(int i) { return &at(i); }

    /* implement Plexer */
    void sigDown(int iCtx, Channel *pcn, int hChan) { pta(hChan)->sigDown(iCtx, pcn); }
    void sigUp(int iCtx, Channel *pcn, int hChan) { pta(hChan)->sigUp(iCtx, pcn); }


   public:
    int init(const std::string &sg, Umi::Channel *pcn);
    MultiPlex(const UmiProto *pup);
    ~MultiPlex();
  };

  /**
   * Dino Protocol Plexer
   */
  class DinoPlex:public MultiPlex,  public Failable {
    struct Bits {
      unsigned char _mpucOut[16];	/* IO bits */
      unsigned char _mpucIn[16];	
      size_t _sIn, _sOut;
      Bits(): _sIn(4), _sOut(4) { 
	memset(_mpucOut, 0, sizeof(_mpucOut)); 
	memset(_mpucIn, 0, sizeof(_mpucIn)); 
      }
      ~Bits() { }
    };
    std::vector<Bits> _vcbi;
    typedef std::vector<Bits>::const_iterator IcBi;
    friend class DinoSyncClientFunc;

   public:
    bool getBit(BitHandle bh) { return _vcbi[ibits(bh)]._mpucOut[index(bh)] & flag(bh); }
    void setBit(BitHandle bh) { _vcbi[ibits(bh)]._mpucOut[index(bh)] |= flag(bh); }
    void clrBit(BitHandle bh) { _vcbi[ibits(bh)]._mpucOut[index(bh)] &= ~flag(bh); }
    void setBit(BitHandle bh, bool b) { return b ? setBit(bh) : clrBit(bh); }

    bool isBit(BitHandle bh) { return _vcbi[ibits(bh)]._mpucIn[index(bh)] & flag(bh); }

    std::ostream &print(std::ostream &os);

    int startOpen(const std::string &sgPort);
    BitHandle doneOpen(int i); 
    void startSync();
    int doneSync();

    int init(const std::string &sg, Umi::Channel *pcn);


    DinoPlex():MultiPlex(&theDino_dino_Proto) { 
    }
  };
  

 
  /**
   * Dino Protocol Plexer with blocking functions
   */
  class DinoPlexSync:public DinoPlex, public Umi::UseIo {
    bool _fRun;

    void done(Umi::Task *pta) ;

   public:
    /**
     * open bit at port. 
     *
     * \param sgPort name of port to open.
     *
     * \return The returned BitHandle can be used as argument for the getBit, setBit, etc. functions. bhNil is returned in case of an error.
     */
    BitHandle callOpen(const std::string &sgPort);

    /**
     * synchronize internal state with field I/O.
     *
     * Read the current input status from the field i/o device, then updates the outputs of the field i/o device.
     */
    int callSync(); 


    DinoPlexSync( Umi::RunIo *pur );
    ~DinoPlexSync(  );
  };
}
#endif

