
/**
 * \file 
 * Rio Functions
 */
#ifndef _RIO_RIO_HXX__
#define _RIO_RIO_HXX__

#ifdef SWIG
%module rio
#endif 

#include <list>
#include <fstream>
#include "umi/umi_task.hxx"
#include "aes/aes.hxx"
#include "rio/protocols.hxx"
#include "aesc/umipro_aes.hxx"

/**
 * Rio namespace
 */
namespace INdepRio100 {	
  namespace Umi=INdepUmi100;
  namespace Aes=INdepAes100 ;

  class RioReadFunc:public Umi::TaskFunc::Func {
    int _cnRio;
    int _iPort;
    unsigned *_pnVal;

    int cnRio() const { return _cnRio; }

   public:
    int func( int eState );
    RioReadFunc(int cnRio, int iPort, unsigned *pnVal):_cnRio(cnRio), _iPort(iPort), _pnVal(pnVal) { assert(pnVal); }
    ~RioReadFunc() { }
  };

  class RioReadnFunc:public RioReadFunc {
    unsigned _nVal;
   public:
    int nVal() const { return _nVal; }
    RioReadnFunc(int cnRio, int iPort):RioReadFunc(cnRio,iPort, &_nVal) {  }
    ~RioReadnFunc() { }
  };

  class RioWriteFunc:public Umi::TaskFunc::Func {
    int _cnRio;
    int _iPort, _iVal;

    int cnRio() const { return _cnRio; }

   public:
    int func( int eState );
    RioWriteFunc(int cnRio, int iPort, int iVal ):_cnRio(cnRio), _iPort(iPort), _iVal(iVal) { }
    ~RioWriteFunc() { }
  };


  class RioConfigFunc:public Umi::TaskFunc::Func,public Failable {
    int _cnRio;
    /* config */
    std::filebuf *_pfb;
    std::stringbuf *_pgb;
    std::streambuf *psb() { return _pfb ? static_cast<std::streambuf*>(_pfb) : static_cast<std::streambuf*>(_pgb); }
    int cnRio() const { return _cnRio; }
    std::string _sgConfig;
   public:
    const std::string &sgConfig() { return _sgConfig; }
    int func( int eState );
    Failable &initPath(const std::list<std::string> &lisgPath, const char *szProg);
    Failable &initFile(const char *szProg);
    Failable &initData(const char *mpch);

    RioConfigFunc(int cnRio);
    ~RioConfigFunc();
  };

  std::list<std::string> getEnv(std::string sgEnv, std::string sgPostfix);

  class RioTaskPlex:public Aes::AesTaskPlex {
    std::string _sgConfig;		/* config loaded */


    Failables &callRioConfig(const char *szFile);
    Failables &callRioConfig(const std::list<std::string>&, int iVers);

   public:
    const std::string &sgConfig() const { return _sgConfig; }

    /** open & config from path list */
    Failables &callRioInitPath(const std::list<std::string>&, int iVers=0);

    /** open & config from specified file */
    Failables &callRioInitFile(const char *szFile);

    /** open & config from char array */
    Failables &callRioInitData(const char *szFile);

    /** reads the register content of iPort */
    int callRioRead(int iPort) { 
      unsigned nVal=-1;
      run(new RioReadFunc(0, iPort, &nVal));
      return nVal;
    }

    /** writes the register content of iPort */
    void callRioWrite(int iPort, int iVal) { 
      run(new RioWriteFunc(0, iPort, iVal));
    }

    RioTaskPlex(Umi::Channel *puc, Umi::ChanUser *pucIo):Aes::AesTaskPlex(puc, pucIo) {
      assert(*(pucIo->pup()) == INdepRio100::theRio_rio_Proto); 
      assert(*(puc->pup()) == INdepRio100::theRio_rio_Proto); 
    }
    ~RioTaskPlex() { }
  };


  /**
   *
   * Convenience class to configure all resources necessary for process-local operation
   * in a single construction operation. Resources remain initialized during the object lifetime and are released during destruction.
   */
  class RioLocalTaskInitializer:public Failables {
    Umi::ChanDevio *_puc;	/* szDevice i/o channel */
    RioTaskPlex *_prp;		/* endpoint plexer */
   public:
    RioTaskPlex *prp() const { return _prp; }

    /**
     * configure all resources necessary for process-local operation
     *
     * \arg szDevice system specific path to device node, e.g. "/dev/itina/0"
     * \arg szRbfPath system specific path to a .rbf file for FPGA initialization
     *
     * On return check via isGood() for successful initialization.
     */
    RioLocalTaskInitializer(const char *szDevice, const char *szRbfPath[]);

    /**
     * Resource allocation, but no FPGA initialization
     *
     * \arg szDevice system specific path to device node, e.g. "/dev/itina/0"
     */
    RioLocalTaskInitializer(const char *szDevice);	/* open only */
    ~RioLocalTaskInitializer();
  };

}


#endif
