
/**
 * \file 
 * Tina Multiplexer
 */
#ifndef _TINA_MUX_HXX__
#define _TINA_MUX_HXX__

#include "cplRaw/flags.hxx"
#include "umi/umi_task.hxx"
#include "rio/protocols.hxx"
#include "tina/protocols.hxx"
#include "dino/dino.hxx"

#define DO_FIFO_MUX 1

namespace INdepDev100 {	
  namespace Umi=INdepUmi100;
  namespace Rio=INdepRio100;
  namespace Dino=INdepDino100;

  typedef int DelE;



  /** 
   * physical device manager 
   */
  class TinaSampleServer;
  class TinaSampleTask;

  class TicMux:public Failables , public Umi::PortMux, public Flags {

   private:
    enum { ceTinaUser, ceTic, ceDino, ccPortTypes };
    static PortClass mpPortClass[ccPortTypes+1];
    class DinoSyncServerFunc;


    /* server ports */
    Port _mpPo[ccPortTypes];

    /* dino map */
    typedef std::map<std::string, std::pair<int,int> > MpSgBi;
    const MpSgBi mp;

    /* config string */
    const char *_szConfig;

    bool isDinoInit() { return eatFlag(0x1)==0; }

    /**
     * create task at port
     */
    Umi::Task *createTask(Umi::Channel *, const PortClass *, Umi::Pmd pmd, Umi::PortMux *pma, int h);
    void deleteTask(const PortClass*, Umi::Task *pta);
  public:

    TicMux( Umi::Channel *pucDrv, const char* szConfig="" );
    ~TicMux();
  };

  /** 
   * fifo manager 
   */
  class FifoMux: public Umi::Plexer {
    Umi::UmiChannel *_pucDrv;
    Umi::UmiChannel *_pucMux;

    Umi::UmiChannel *pucDrv() { return _pucDrv; }	/* to device hardware */
    Umi::UmiChannel *pucMux() { return _pucMux; }	/* to device user */


  public:
    void sigDown(int iCtx, Umi::Channel *pcn, int hChan);
    void sigUp(int iCtx, Umi::Channel *pcn, int hChan);


    FifoMux( Umi::UmiChannel *pucDrv, Umi::UmiChannel *pucMux );
    ~FifoMux();
  };

}


#endif
