
#ifndef _REGS_HXX_
#define _REGS_HXX_ 1


#if (MMSYS_wdm)
extern "C" { 
  /*
   * WinNT Kernel specifics
   */
  #include <wdm.h>
} 
#endif

#if (MMSYS_qnx)
  #include <x86/inout.h>
#endif

#if (MMSYS_lkern)
  #if defined(__i386__) || defined(__x86_64__)
    /*
     * Linux Kernel specifics
     */
    static __inline unsigned char inb (unsigned short int port)
    {
      unsigned char _v;

      __asm__ __volatile__ ("inb %w1,%0":"=a" (_v):"Nd" (port));
      return _v;
    }

    static __inline void outb (unsigned char value, unsigned short int port)
    {
      __asm__ __volatile__ ("outb %b0,%w1": :"a" (value), "Nd" (port));
    }
  #endif
  #ifdef __powerpc__

    /*
     * from ~/linux-2.4.25/include/asm-ppc/io.h
     *
     * 8, 16 and 32 bit, big and little endian I/O operations, with barrier.
     *
     * Read operations have additional twi & isync to make sure the read
     * is actually performed (i.e. the data has come back) before we start
     * executing any following instructions.
     */
    extern inline int in_8(volatile unsigned char *addr)
    {
	    int ret;

	    __asm__ __volatile__(
		    "lbz%U1%X1 %0,%1;\n"
		    "twi 0,%0,0;\n"
		    "isync" : "=r" (ret) : "m" (*addr));
	    return ret;
    }

    extern inline void out_8(volatile unsigned char *addr, int val)
    {
	    __asm__ __volatile__("stb%U0%X0 %1,%0; eieio" : "=m" (*addr) : "r" (val));
    }
    extern inline unsigned inb(unsigned int port) { return in_8((unsigned char *)port); }
    extern inline void outb(int val, unsigned int port) { out_8((unsigned char *)port, val); }

  #endif


  extern int printk( const char *fmt, ...);
  #define PREFIX "<0>core488: "
#endif

  /*
   * Registers
   */

namespace INdepUmi100 {
  class Regs {
    int _ioBase;

  /*
   * kernel space code - direct i/o
   */
  #if (MMSYS_wdm)	/* WinNT kernel */
   public:
    /* for WDM driver */
    int read(int iReg) { return READ_PORT_UCHAR((PUCHAR)_ioBase+iReg); }
    //int readByte(int iReg) { return READ_PORT_UCHAR((PUCHAR)_ioBase+iReg); }
    void write(int iReg, int i) { WRITE_PORT_UCHAR((PUCHAR)_ioBase+iReg, (UCHAR)i); }
    bool isGood() const { return true; }

    Regs(int ioBase):_ioBase(ioBase) {}
    ~Regs() {}

  #endif 

    

  #if (defined (MMSYS_lkern) && (!defined(MMISA_mips)))
   public:
    bool isGood() const { return true; }
    int read(int iReg) { return inb(_ioBase+iReg); }
    //int readByte(int iReg) { return inb(_ioBase+iReg); }
    void write(int iReg, int i) { outb(i, _ioBase+iReg); }

    Regs(int ioBase):_ioBase(ioBase) {}
    ~Regs() {}

  #endif 

  #if (defined(MMSYS_qnx))	
   public:
    bool isGood() const { return true; }
    int read(int iReg) { return in8(_ioBase+iReg); }
    //int readByte(int iReg) { return in8(_ioBase+iReg); }
    void write(int iReg, int i) { out8(_ioBase+iReg, i); }

    Regs(int ioBase):_ioBase(ioBase) {}
    ~Regs() {}

  #endif 

  /*
   * user space code - allow overload to enable redirection to virtual i/o 
   */
  #if (defined(MMSYS_linux) || defined(MMSYS_mingw) || MMSYS_win || (defined(MMSYS_lkern) && defined(MMISA_mips)))
    
    void *_pMemIo;
    volatile void *pMemIo() { return static_cast<volatile void *>(_pMemIo); }


   public:
    virtual bool isGood() const { return _pMemIo != 0; }
    virtual int read(int iReg) { return static_cast<volatile int *>(pMemIo())[iReg] & 0xff; }
    virtual void write(int iReg, int i) { static_cast<volatile int *>(pMemIo())[iReg] = i; }
    //virtual int readByte(int iReg) { return static_cast<volatile unsigned char *>(pMemIo())[iReg] & 0xff; }

    //Regs(int iAddrMem);
    //~Regs();

    Regs(int ioBase):_pMemIo((void*) ioBase) {}
    ~Regs() {}
  #endif 
    int addr() const { return _ioBase; }

  };
} 


#endif
