#include <linux/types.h>
#include "sizes.h"
#include <linux/isa.h>
#include "umic/umiad.h"
#include "card.h"
#include "mod.h"
#include "assert.h"
#include <linux/sched.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <linux/interrupt.h>

 /*
  * usb.c
  *
  * driver for USB adaptors
  */
  #define SZ_USBSTAT MAX_CACN

  #define MAX_CUSB 2		/* max channels per driver USB limit		*/
  #define MAX_CURB 4		/* max urb per channel				*/

  struct CardUsbS;
  typedef struct CardUsbS CardUsb;


  /*
   * URB associated resources
   */
  struct UrbRscS {		
    void *_pBuf;
    size_t _sBuf;
    dma_addr_t _dma;
    struct urb *_pUrb;
  };
  typedef struct UrbRscS UrbRsc;

  /*
   * 2 usb endpoints per Umi channel (put and get)
   */
  struct ChanUsbS {
    CardUsb *_pcu;

    /* wait queues */
    wait_queue_head_t _wqPut;	/* waiting to put message downstream	*/
    wait_queue_head_t _wqGet;	/* waiting to get message upstream	*/

    atomic_t _eStPut;		/* put s/m */
    atomic_t _eStGet;		/* get s/m */

    size_t _sepPut;	/* put endpoint */
    size_t _sepGet;	/* put endpoint */
    int _iepPut;	/* put endpoint */
    int _iepGet;	/* get endpoint */


    UrbRsc _mpur[MAX_CURB];
  };
  #define pcn_pcu(p) ((p)->_pcu)

  #define pcn_pisidle(p) (atomic_read(&(p)->_eStPut) == 0)	/* no urb scheduled    */
  #define pcn_pisctrl(p) (atomic_read(&(p)->_eStPut) == 1)	/* control in progress */
  #define pcn_pisdata(p) (atomic_read(&(p)->_eStPut) == 2)	/* data in progress    */

  #define pcn_psetidle(p) (atomic_set(&(p)->_eStPut, 0))
  #define pcn_psetctrl(p) (atomic_set(&(p)->_eStPut, 1))
  #define pcn_psetdata(p) (atomic_set(&(p)->_eStPut, 2))

  #define pcn_gisidle(p) (atomic_read(&(p)->_eStGet) == 0) 	/* no urb scheduled    */
  #define pcn_gisdata(p) (atomic_read(&(p)->_eStGet) == 1) 	/* waiting for data    */
  #define pcn_gisctrl(p) (atomic_read(&(p)->_eStGet) == 2) 	/* waiting for control */
  #define pcn_gisdone(p) (atomic_read(&(p)->_eStGet) == 3) 	/* ctrl and data available  */

  #define pcn_gsetidle(p) (atomic_set(&(p)->_eStGet,0)) 	
  #define pcn_gsetdata(p) (atomic_set(&(p)->_eStGet,1))
  #define pcn_gsetctrl(p) (atomic_set(&(p)->_eStGet,2)) 
  #define pcn_gsetdone(p) (atomic_set(&(p)->_eStGet,3)) 

  #define pcn_pcur(p) (&((p)->_mpur[0]))	/* put/control UrbRsc */
  #define pcn_pdur(p) (&((p)->_mpur[1]))	/* put/data UrbRsc */
  #define pcn_gcur(p) (&((p)->_mpur[2]))	/* get/control UrbRsc */
  #define pcn_gdur(p) (&((p)->_mpur[3]))	/* get/data UrbRsc */
  typedef struct ChanUsbS ChanUsb;

  /* per channel status flags */

  struct CardUsbS {
    Card _ca;				/* header			*/

    /*
     * Adaptor information
     */
    UmiAdaptUsb *_pad;
    int _fOpen;				/* open flag			*/

    struct usb_device *_pud;		
    ChanUsb _mpcn[MAX_CUSB]; 		/* channels */

  } ;

  #define setBusy(p) 		/* set/clr busy flag	*/
  #define clrBusy(p) 
  #define pcu_pca(p) ((&(p)->_ca))
  #define pcu_pdm(p) (pca_pdm(&(p)->_ca))
  #define pcu_pcn(p, i) ((p)->_mpcn+(i))
  #define pcu_sep(p, i, m) ((p)->_mpcn[(i)]._sep##m)
  #define pcu_iep(p, i, m) ((p)->_mpcn[(i)]._iep##m)


  /* 
   * firmware loader
   */
  struct ezusb_hex_record {
	  __u16 _addr;
	  __u8 _cData;
	  __u8 _data[16];
  };

  /*
   * ezusb_writememory ()
   */
  int ezusb_writememory (struct usb_device *pud, int addr, unsigned char *pchData, int length, __u8 bRequest) {
    int result;
    unsigned char *pch;

    /* dbg("ezusb_writememory %x, %d", addr, length); */

    pch =  kmalloc (length, GFP_KERNEL);
    if (!pch) {
      //printk(KERN_ERR PFX "kmalloc(%d) failed\n", pcu_pdm(pcu)->_szDid, length);
      return -ENOMEM;
    }
    memcpy (pch, pchData, length);
    result = usb_control_msg (pud, usb_sndctrlpipe(pud, 0), bRequest, 0x40, addr, 0, pch, length, 3*HZ);
    //printk(KERN_DEBUG "result(%d)\n", result);
    kfree (pch);
    return result;
  }



  /*
   * ezusb_set_reset()
   */
  int ezusb_set_reset (struct usb_device *pud, unsigned char reset_bit) {
    int	response;
    /* write CPUCS register */
    response = ezusb_writememory (pud, 0xe600, &reset_bit, 1, 0xa0);
    return response;
  }

  #define isDebugUsb(pcu) (isDebug(pcu_pdm(pcu), &(pcu)->_ca))

  void usbPutComplete(struct urb *pUrb) {
    ChanUsb *pcn=(ChanUsb *)pUrb->context;

    /* sync/async unlink faults aren't errors */
    if (pUrb->status &&
      !(pUrb->status == -ENOENT ||
        pUrb->status == -ECONNRESET ||
        pUrb->status == -ESHUTDOWN)) {
	  printk(KERN_ERR PFX "nonzero write bulk status received: %d", pcu_pdm(pcn_pcu(pcn))->_szDid, pUrb->status);
    }

    if (pUrb == pcn_pcur(pcn)->_pUrb ) {
      assert(pcn_pisctrl(pcn));
      pcn_psetdata(pcn);
      wake_up_interruptible(&pcn->_wqPut);
    }
    else if (pUrb == pcn_pdur(pcn)->_pUrb ) {
      /* done */
      assert(pcn_pisdata(pcn));
      pcn_psetidle(pcn);
      wake_up_interruptible(&pcn->_wqPut);
    }
    else assert(0);
  }

#if 0
  ChanUsb *pcn_pcu(struct CardStruct *pca, struct ChanStruct*pcc) { 
    int iChan=pcc->_iChanOfCard;
    CardUsb *pcu=(CardUsb*)pca;
    ChanUsb *pcn=pcu_pcn(pcu, iChan);
    assert(iChan >= 0 && iChan < MAX_CUSB);

    return pcn;
  }
#endif


  /*
   * put down
   */
  int usbPutDown(Card *pca, Chan *pcc, UmiIoc *pubU) {
    int iChan=pcc->_iChanOfDriver;
    int e=0;
    CardUsb *pcu=(CardUsb*)pca;
    ChanUsb *pcn=pcu_pcn(pcu, iChan);
    assert(iChan >= 0 && iChan < MAX_CUSB);


    if (1) {
      size_t c;
      /* new implementation */

      /* wait here until ready */
      e = wait_event_interruptible(pcn->_wqPut, (pcn_pisidle(pcn)));
      if (e) return e;

      pcn_psetctrl(pcn);

      if (1) {
	char mpch[SZ_CTRL_MAX];
	char *pchBuf=pcn_pcur(pcn)->_pBuf;
	size_t sBuf=pcn_pcur(pcn)->_sBuf;

	assert(pchBuf);
	assert(sBuf);
	assert(pcn);

	/* copy down the control message into temporary buffer */
	if ((e=bufPutDown(&pubU->_cd, mpch, SZ_CTRL_MAX, &c))) return e;
	assert(c > 0);	/* control never zero length */

	/* marshal the data into the transfer buffer */
	c=UmiProto_marshal(pcc->_pUdd->_pup, pchBuf, sBuf, (struct UmiControlStruct *)mpch);

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "usb put down ctrl, chan=%d, size=%zu h=%d\n", pcu_pdm(pcu)->_szDid, pcc->_iChanOfDriver, c, *((unsigned char*)pchBuf));



        assert( pcn_pcur(pcn)->_pUrb);
        assert( pcc );

	/* initialize the control urb properly */
	usb_fill_bulk_urb(
	  pcn_pcur(pcn)->_pUrb, 
	  pcu->_pud, 
	  usb_sndbulkpipe(pcu->_pud, pcu_iep(pcu,pcc->_iChanOfDriver,Put)),
	  pchBuf, 
	  c,
	  usbPutComplete,
	  pcn
	);
	pcn_pcur(pcn)->_pUrb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "submit urb put down ctrl\n", pcu_pdm(pcu)->_szDid);

	e=usb_submit_urb(pcn_pcur(pcn)->_pUrb, GFP_KERNEL);
	if (e) {
	  printk(KERN_ERR PFX "usb can't submit ctrl put urb (%d)\n", pcu_pdm(pcu)->_szDid, e);
	  pcn_psetidle(pcn);	/* safe: complete never called */
	  return e;
	}

	e = wait_event_interruptible(pcn->_wqPut, (pcn_pisdata(pcn)));

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "transfer done %d %d\n", pcu_pdm(pcu)->_szDid, e, pcn_pcur(pcn)->_pUrb->status);
      }

      if (1) {

	/* copy down the data message */
	assert(pcn_pdur(pcn)->_pBuf);
	if ((e=bufPutDown(&pubU->_dd, pcn_pdur(pcn)->_pBuf, pcn_pdur(pcn)->_sBuf, &c))) return e;


	assert(pcn_pdur(pcn)->_pUrb);

	if (c==0) {
	  /* REMIND: can't send zero length message */
	  memset(pcn_pdur(pcn)->_pBuf, 0, 2);
	  c=2;
	}
	if (isDebugUsb(pcu)) printk(KERN_DEBUG PFX "usb put down data size=%zu\n", pcu_pdm(pcu)->_szDid, c);

	/* initialize the data urb properly */
	usb_fill_bulk_urb(
	  pcn_pdur(pcn)->_pUrb, 
	  pcu->_pud, 
	  usb_sndbulkpipe(pcu->_pud, pcu_iep(pcu,pcc->_iChanOfDriver,Put)),
	  pcn_pdur(pcn)->_pBuf, 
	  c,
	  usbPutComplete,
	  pcn
	);
	pcn_pdur(pcn)->_pUrb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	if (isDebugUsb((pcu))) 
	  printk(KERN_DEBUG PFX "submit urb put down data\n", pcu_pdm(pcu)->_szDid);

	e=usb_submit_urb(pcn_pdur(pcn)->_pUrb, GFP_KERNEL);
	if (e) {
	  printk(KERN_ERR PFX "usb can't submit data put urb (%d)\n", pcu_pdm(pcu)->_szDid, e);
	  pcn_psetidle(pcn);	/* safe: control complete does not touch flags */
	  return e;
	}

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "submit done\n", pcu_pdm(pcu)->_szDid);

	e = wait_event_interruptible(pcn->_wqPut, (pcn_pisidle(pcn)));

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "transfer done %d %d\n", pcu_pdm(pcu)->_szDid, e, pcn_pdur(pcn)->_pUrb->status);
      }
    }
    return e;
  }

  void usbGetComplete(struct urb *pUrb) {
    ChanUsb *pcn=(ChanUsb *)pUrb->context;

    /* sync/async unlink faults aren't errors */
    if (pUrb->status &&
      !(pUrb->status == -ENOENT ||
        pUrb->status == -ECONNRESET ||
        pUrb->status == -ESHUTDOWN)) {
	  printk(KERN_ERR PFX "nonzero receive bulk status received: %d", pcu_pdm(pcn_pcu(pcn))->_szDid, pUrb->status);
    }

    if (!pcn_gisidle(pcn)) {
      if (pUrb == pcn_gdur(pcn)->_pUrb ) {
	assert(pcn_gisdata(pcn));
	pcn_gsetctrl(pcn);
	wake_up_interruptible(&pcn->_wqGet);
      }
      else if (pUrb == pcn_gcur(pcn)->_pUrb ) {
	/* done */
	assert(pcn_gisctrl(pcn));	/* \TRIG 2014-10-29-00 */
	pcn_gsetdone(pcn);
	wake_up_interruptible(&pcn->_wqGet);
      }
      else assert(0);
    }
  }

  int usbGetStart(Chan *pcc, CardUsb *pcu, ChanUsb *pcn) {
    int e;

    if (pcn_gisidle(pcn)) {
      /* initialize transfers */
      pcn_gsetdata(pcn);

      if (1) {
	/* get data first */
	/* initialize the data urb properly */

	usb_fill_bulk_urb(
	  pcn_gdur(pcn)->_pUrb, 
	  pcu->_pud, 
	  usb_rcvbulkpipe(pcu->_pud, pcu_iep(pcu,pcc->_iChanOfDriver,Get)),
	  pcn_gdur(pcn)->_pBuf, 
	  pcn_gdur(pcn)->_sBuf, 
	  usbGetComplete,
	  pcn
	);
	pcn_gdur(pcn)->_pUrb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "submit urb get data \n", pcu_pdm(pcu)->_szDid);

	e=usb_submit_urb(pcn_gdur(pcn)->_pUrb, GFP_KERNEL);
	if (e) {
	  printk(KERN_ERR PFX "usb can't submit get data urb (%d)\n", pcu_pdm(pcu)->_szDid, e);
	  pcn_gsetidle(pcn);	
	  return e;
	}

	//e = wait_event_interruptible(pcn->_wqGet, (pcn_gisctrl(pcn)));
	//if (isDebugUsb(pcu)) printk(KERN_DEBUG PFX "transfer done %d %d\n", pcu_pdm(pcu)->_szDid, e, pcn_gdur(pcn)->_pUrb->status);
      }
      
      if (1) {
	/* get control */

	usb_fill_bulk_urb(
	  pcn_gcur(pcn)->_pUrb, 
	  pcu->_pud, 
	  usb_rcvbulkpipe(pcu->_pud, pcu_iep(pcu,pcc->_iChanOfDriver,Get)),
	  pcn_gcur(pcn)->_pBuf, 
	  pcn_gcur(pcn)->_sBuf, 
	  usbGetComplete,
	  pcn
	);
	pcn_gcur(pcn)->_pUrb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	if (isDebugUsb(pcu)) 
	  printk(KERN_DEBUG PFX "submit urb get ctrl \n", pcu_pdm(pcu)->_szDid);

	e=usb_submit_urb(pcn_gcur(pcn)->_pUrb, GFP_KERNEL);
	if (e) {
	  printk(KERN_ERR PFX "usb can't submit get ctrl urb (%d)\n", pcu_pdm(pcu)->_szDid, e);
	  pcn_gsetidle(pcn);	
	  return e;
	}

	//e = wait_event_interruptible(pcn->_wqGet, (pcn_gisdone(pcn)));
	//if (isDebugUsb(pcu)) printk(KERN_DEBUG PFX "transfer done %d %d\n", pcu_pdm(pcu)->_szDid, e, pcn_gcur(pcn)->_pUrb->status);
      }
    }
    return 0;
  }

  int usbGetUp(Card *pca, Chan *pcc, UmiIoc *pubU) {
    int iChan=pcc->_iChanOfDriver;
    CardUsb *pcu=(CardUsb*)pca;
    ChanUsb *pcn=pcu_pcn(pcu, iChan);
    register int e=0;
    assert(iChan >= 0 && iChan < MAX_CACN);

    if (1) {
      /* new implementation */
      if ( (e=usbGetStart(pcc, pcu, pcn)) ) return e;

      /* wait here until data are available */
      e = wait_event_interruptible(pcn->_wqGet, (pcn_gisdone(pcn)));
      /*
       * REMIND 
       * in case of a signal we might return here without pcn_gisdone true (outstanding URBs) and should drain/reset the USB .
       * assert(pcn_gisctrl(pcn)) triggered in usbGetComplete() when restarting after interrupting a user process.
       */
      pcn_gsetidle(pcn);
      if (e) return e;

      if (1) {
	size_t c;
	/* copy data part to the user */
	e=bufGetUp(&pubU->_du, pcn_gdur(pcn)->_pBuf, c=pcn_gdur(pcn)->_pUrb->actual_length);
	if (e) { 
	  printk(KERN_ERR PFX "can't copy up %zu %d\n", pcu_pdm(pcu)->_szDid, c,e);
	  return e;
	}
	if (isDebugUsb(pcu)) printk(KERN_DEBUG PFX "usb getup data size=%zu\n", pcu_pdm(pcu)->_szDid, c);
      }


      if (1) {
	size_t c;
	char *pchBuf=pcn_gcur(pcn)->_pBuf; 
	size_t l=pcn_gcur(pcn)->_pUrb->actual_length;
	char mpch[SZ_CTRL_MAX];

	if (isDebugUsb(pcu)) {
	  int i;
	  printk(KERN_DEBUG PFX "ctrl ", pcu_pdm(pcu)->_szDid);
	  for (i=0; i<l; i++) {
	    printk("%02x ", (((unsigned char *)pchBuf)[i]));
	  }
	  printk("\n");
	}

	/* unmarsh the packet to local buffer */
	c=UmiProto_unmarsh(pcc->_pUdd->_pup, (struct UmiControlStruct *)mpch, pchBuf, l);
	if (c == 0) {
	  printk(KERN_ERR PFX "can't unmarsh h=%d l=%zd\n", pcu_pdm(pcu)->_szDid, *((unsigned char*)pchBuf),l);
	  return -EPROTO;
	}

	/* copy data part to the user */
	e=bufGetUp(&pubU->_cu, mpch, c);
	if (e) { 
	  printk(KERN_ERR PFX "can't copy up %zu %d\n", pcu_pdm(pcu)->_szDid, c,e);
	  return e;
	}
	if (isDebugUsb(pcu)) printk(KERN_DEBUG PFX "usb getup ctrl size=%zu\n", pcu_pdm(pcu)->_szDid, c);
      }

      /* restart receive */
      e=usbGetStart(pcc, pcu, pcn);
    }
    return e;
  }

  int usbDownload(CardUsb *pcu, const struct ezusb_hex_record *phr) {
    int e;
    assert(phr);

    /* firmware download */
    if ((e=ezusb_set_reset (pcu->_pud, 1))<0) {
      printk(KERN_ERR PFX "usb can't reset (%d)\n", pcu_pdm(pcu)->_szDid, e);
      return -1;
    }
    while(phr->_addr != 0xffff) {
      e = ezusb_writememory(pcu->_pud, phr->_addr, (unsigned char *)phr->_data, phr->_cData, 0xa0);
      if (e < 0) {
	printk(KERN_ERR PFX "usb can't load firmware (%d %04X %p %d)\n", pcu_pdm(pcu)->_szDid, e, phr->_addr, phr->_data, phr->_cData);
	return -1;
      }
      phr++;
    }
    if ((e=ezusb_set_reset (pcu->_pud, 0))<0) {
      printk(KERN_ERR PFX "usb can't restart (%d)\n", pcu_pdm(pcu)->_szDid, e);
      return -1;
    }
    printk(KERN_INFO PFX "usb download done\n", pcu_pdm(pcu)->_szDid);
    return 0;
  }

  int usbInit(Card *pca, int iDriver) {
    CardUsb *pcu=(CardUsb*)pca;

    if (pcu->_fOpen == 0) {
      const struct ezusb_hex_record *phr=pcu->_pad->_mppfw[iDriver];
      /* first open of any channel */
      if (!phr) return -ENODEV;

      if (usbDownload(pcu, phr)) return -ENODEV;

      /* delay for init the ezusb to start and initialize */
      udelay(1000);
    }

    pcu->_fOpen |= 1;
    return 0;
  }

  void usbFini(Card *pca, int iDriver) {
    CardUsb *pcu=(CardUsb*)pca;

    pcu->_fOpen = 0;

    if (pcu->_fOpen == 0) {
      /*
       * load idle firmware
       */
      if (pcu->_pad->_pfwIdle) usbDownload(pcu, pcu->_pad->_pfwIdle);
    }
  }

  char* usbPrint( Card *pca, char* pchBuf, int c) {
    return pchBuf;
  }


  /*
   * Card construction/destruction
   */
  CardUsb *usbCardDelete(LumiMod *pdm, CardUsb *pcu) {
    int iChan;
    ChanUsb *pcn;

    if (pcu) {
      for ( pcn=pcu->_mpcn, iChan=0; iChan<MAX_CUSB; iChan++, pcn++ ) {
	int iUrb;
	UrbRsc *pur;
	for ( iUrb=0, pur=pcn->_mpur; iUrb < MAX_CURB; iUrb++, pur++ ) {
	  if (pur->_pBuf) usb_free_coherent(pcu->_pud, pur->_sBuf, pur->_pBuf, pur->_dma);
	  if (pur->_pUrb) usb_free_urb(pur->_pUrb);
	}

      }
      kfree (pcu);
    }
    return 0;
  }

  unsigned int usbPoll(struct CardStruct *pca, struct ChanStruct*pcc, struct file *fp, poll_table *pt ) { 
    int f=0;
    int iChan=pcc->_iChanOfDriver;
    CardUsb *pcu=(CardUsb*)pca;
    ChanUsb *pcn=pcu_pcn(pcu, iChan);
    assert(iChan >= 0 && iChan < MAX_CUSB);

    if (pcn_gisidle(pcn)) {
      usbGetStart(pcc, pcu, pcn);	/* no way to return error, log is already provided by usbGetStart() */
    }
    poll_wait(fp, &pcn->_wqPut, pt);
    poll_wait(fp, &pcn->_wqGet, pt);

    if (pcn_pisidle(pcn)) f|=POLLOUT|POLLWRNORM;
    if (pcn_gisdone(pcn)) f|=POLLIN|POLLRDNORM;
    return f; 
  }

  CardUsb *usbCardNew(LumiMod *pdm, struct usb_device *pud, UmiAdaptUsb *pad ) {
    int iChan;
    ChanUsb *pcn;
    CardUsb *pcu=(CardUsb*)kmalloc(sizeof(CardUsb), GFP_KERNEL);
    if (!pcu) return usbCardDelete(pdm, pcu);

    pca_init(&pcu->_ca, pdm );
    pcu->_pud=pud;
    pcu->_pad=pad;
    pcu->_ca._pfPutDown=usbPutDown;
    pcu->_ca._pfGetUp=usbGetUp;
    pcu->_ca._pfPoll=usbPoll;
    pcu->_ca._pfInit=usbInit;
    pcu->_ca._pfFini=usbFini;
    pcu->_ca._pfPrint=usbPrint;
    pcu->_fOpen=0;
    memset(pcu->_mpcn, 0, sizeof(pcu->_mpcn));

    for ( pcn=pcu->_mpcn, iChan=0; iChan<MAX_CUSB; iChan++, pcn++ ) {
      UrbRsc *pur;
      int iUrb;

      pcn->_pcu=pcu;
      atomic_set(&pcn->_eStPut,0);
      atomic_set(&pcn->_eStGet,0);
      pcn->_iepPut=pcn->_iepGet=-1;
      pcn->_sepGet=pcn->_sepPut=0;

      for ( iUrb=0, pur=pcn->_mpur; iUrb < MAX_CURB; iUrb++, pur++ ) {
	pur->_pUrb=usb_alloc_urb(0, GFP_KERNEL);
	if (!pur->_pUrb) return usbCardDelete(pdm, pcu);

	pur->_pBuf=usb_alloc_coherent(pcu->_pud, pur->_sBuf=max( SZ_CTRL_MAX, SZ_DATA_MAX), GFP_KERNEL, &pur->_pUrb->transfer_dma);
	pur->_dma=pur->_pUrb->transfer_dma;
	if (!pur->_pBuf) return usbCardDelete(pdm, pcu);
      }

      init_waitqueue_head(&pcn->_wqPut);
      init_waitqueue_head(&pcn->_wqGet);
    }
    return pcu;
  }



  static char * szIdentUsb( char mzBuf[64], struct usb_device *pud ) {
    sprintf(mzBuf, "usb:");
    usb_make_path(pud, mzBuf+4, 60);
    return mzBuf;
  }




  int pdm_usb_probe(LumiMod *pdm, struct usb_interface *pui, const struct usb_device_id *pdi) {
    int i;
    int iSlot;
    CardUsb *pcu;
    struct device *pcd;
    struct usb_device *pud=interface_to_usbdev(pui);
    char mzPath[64];
    const char *szId=szIdentUsb(mzPath, pud);
    struct usb_host_interface *pUhi;
    struct usb_endpoint_descriptor *pUed;

    printk(KERN_INFO PFX "usb probe for %x:%x at %s\n", pdm->_szDid, pdi->idVendor, pdi->idProduct, szId);
    printk(KERN_INFO PFX "usb num_altsetting %d\n", pdm->_szDid, pui->num_altsetting);

    /* select altsetting to scan */
    if (pui->num_altsetting < 4) {
      printk(KERN_ERR PFX "invalid interface\n", pdm->_szDid);
      return -1;
    }
    pUhi = pui->cur_altsetting+1;

    if (pUhi->desc.bNumEndpoints <5) {
      printk(KERN_ERR PFX "wrong config\n", pdm->_szDid);
      return -1;
    }

    for (i = 0; i < pUhi->desc.bNumEndpoints; i++) {
      pUed = &pUhi->endpoint[i].desc;
      printk(KERN_INFO PFX " ep=%d attr=%x addr=%x bs=%d\n", pdm->_szDid, i, pUed->bmAttributes, pUed->bEndpointAddress, pUed->wMaxPacketSize);
    }

    if (usb_set_interface(pud, 0, 1)) {
      printk(KERN_ERR PFX "cant set interface/alt\n", pdm->_szDid);
      return -1;
    }


    /* get s slot */
    if ((iSlot=pdm_getslot(pdm, szId))==-1) return -1;

    /*
     * allocate and initialize a card structure
     */
    pcu=usbCardNew(pdm,pud, (UmiAdaptUsb*)pdi->driver_info);
    if (!pcu) return -1;


    /* 
     * configure pipes 
     * we have 2 bidir pipes, assigned to chan 0 and 1 of the driver
     */
    pcu->_mpcn[0]._iepPut = pUhi->endpoint[2].desc.bEndpointAddress;
    pcu->_mpcn[1]._iepPut = pUhi->endpoint[3].desc.bEndpointAddress;
    pcu->_mpcn[0]._iepGet = pUhi->endpoint[4].desc.bEndpointAddress;
    pcu->_mpcn[1]._iepGet = pUhi->endpoint[5].desc.bEndpointAddress;

    pcu->_mpcn[0]._sepPut = pUhi->endpoint[2].desc.wMaxPacketSize;
    pcu->_mpcn[1]._sepPut = pUhi->endpoint[3].desc.wMaxPacketSize;
    pcu->_mpcn[0]._sepGet = pUhi->endpoint[4].desc.wMaxPacketSize;
    pcu->_mpcn[1]._sepGet = pUhi->endpoint[5].desc.wMaxPacketSize;

    /*
     * register the card
     */
    pdm_setcard(pdm, iSlot, &pcu->_ca);

    /*
     * register the device 
     *
     * first parameter defines the class in sysfs
     * second parameter defines the parent device in sysfs
     */
    pcd=device_create(pdm->_pcl,&pui->dev,MKDEV(pdm->_iMajor,iSlot),&pui->dev,"%s%d", pdm->_szDid, iSlot);
    if (IS_ERR(pcd)) {
      printk(KERN_ERR PFX "device_create failed (%ld), %p\n", pdm->_szDid, PTR_ERR(pcd), pdm->_pcl);
    }

    if (pcu->_pad->_pfwIdle) usbDownload(pcu, pcu->_pad->_pfwIdle);

    return 0;
  }

  /*
   * Note: card deletion must be locked against any use of the card
   */
  void pdm_usb_disconnect(LumiMod*pdm, struct usb_interface *pui) {
    struct usb_device *pud=interface_to_usbdev(pui);
    char mpch[SZ_CARD_ID];

    int iSlot;
    if ((iSlot=pdm_findslot(pdm, szIdentUsb(mpch, pud)))!=-1) {
      CardUsb*pcu;
      if ((pcu=(CardUsb*)pdm_clrcard(pdm, iSlot))) {
	usbDownload(pcu, pcu->_pad->_pfwDiscon);	/* ignore any error */
	printk(KERN_INFO PFX "usb release adaptor %s@%d\n", pdm->_szDid, mpch, iSlot);
	device_destroy(pdm->_pcl,MKDEV(pdm->_iMajor, iSlot));
	printk(KERN_INFO PFX "usb delete card\n", pdm->_szDid);
	usbCardDelete(pdm, pcu);
	printk(KERN_INFO PFX "usb disconnect done\n", pdm->_szDid);
      }
    }
  }
